/*
 * hotplug.h
 *
 * A version of /sbin/hotplug that is not a script.
 *
 * Why?  Think initrd :)
 *
 * Copyright (C) 2001 Greg Kroah-Hartman <greg@kroah.com>
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 * 
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 * 
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef HOTPLUG_H
#define HOTPLUG_H

// If we define this, hotplug will wait "/lib/modules/2.6.34-VENUS/kernel" to be ready.
#define WAIT_MODULE_DIR

#define ADD_STRING	"add"
#define REMOVE_STRING	"remove"
#define CHANGE_STRING	"change"
#define LINKUP_STRING	"linkup"
#define LINKDOWN_STRING	"linkdown"
#define NET_PBC_STRING	"NET_PBC"
#define USB_AI_UP	"USB_AI_UP"
#define USB_AI_DOWN	"USB_AI_DOWN"
#define USB_AO_UP	"USB_AO_UP"
#define USB_AO_DOWN	"USB_AO_DOWN"
#define USB_VI_UP	"USB_VI_UP"
#define USB_VI_DOWN	"USB_VI_DOWN"
#define USB_VO_UP	"USB_VO_UP"
#define USB_VO_DOWN	"USB_VO_DOWN"
#define USB_KBD_UP	"USB_KBD_UP"
#define USB_KBD_DOWN	"USB_KBD_DOWN"
#define USB_MICE_UP	"USB_MICE_UP"
#define USB_MICE_DOWN	"USB_MICE_DOWN"
#define USB_MT_UP	"USB_MT_UP"
#define USB_MT_DOWN	"USB_MT_DOWN"
#define USB_JOYSTICK_UP		"USB_JOYSTICK_UP"
#define USB_JOYSTICK_DOWN	"USB_JOYSTICK_DOWN"
#define SCSI_OFFLINE	"SCSI_OFFLINE"
#define USB_CABLE_BAD	"USB_CABLE_BAD"
#define USB_OVERCUR	"over-current"

#define HOTPLUG_CONFIG_FILE "/var/lock/hotplug/config"
#define HOTPLUG_MOUNT_TMP "/var/lock/hotplug/mount_tmp"
#define HOTPLUG_CONVERT_TMP "/var/lock/hotplug/convert_tmp"

struct subsystem {
	const char * name;
	int (* handler) (void);
};

#ifdef DEBUG
#include <syslog.h>
	#define dbg(format, arg...) do { log_message (LOG_DEBUG, "__FUNCTION__: " format, ## arg); } while (0)
#else
	#define dbg(format, arg...) do { } while (0)
#endif

extern int log_message (int level, const char *fmt, ...)  __attribute__ ((__format__(__printf__, 2, 3)));


extern int split_3values (const char *string, int base, unsigned int * value1, unsigned int * value2, unsigned int * value3);
extern int split_2values (const char *string, int base, unsigned int * value1, unsigned int * value2);
extern int call_subsystem (const char *string, struct subsystem *subsystem);
extern int load_module (const char *module_name);
extern int unload_module (const char *module_name);
//extern int signal_pid(const char *file, int signal);
extern int parse_config(const char *file, const char *name, const char *target, char *value);
extern void wait_dir_ready(const char *name, int times);
extern int create_lockdata(const char *filename);
extern int write_lockdata(const char *filename, const char *data, int len);
extern int write_lockdata_fd(int fd, const char *data, int len);
extern int read_lockdata(const char *filename, char *data, int *len);
extern int testwait_lockdata(const char *filename, const char *data, int len);
extern int testclear_lockdata(const char *filename, const char *data, int len);
extern int testclear_lockdata_once(const char *filename);

extern int usb_handler (void);
extern int scsi_handler (void);
extern int block_handler (void);
extern int firmware_handler (void);
extern int net_handler (void);
extern int hid_handler (void);
extern int pci_handler (void);
//extern int ieee1394_handler (void);

#endif

